from loopy.target import (TargetBase,
                          ASTBuilderBase,
                          DummyHostASTBuilder,
                          )
from loopy.target.c import CASTBuilder
from loopy.target.c.codegen.expression import ExpressionToCExpressionMapper, CExpressionToCodeMapper

_registry = {'float32': 'float',
             'int32': 'int',
             'float64': 'double',
             'string': 'std::string',
             'str': 'std::string'}


class MyMapper(ExpressionToCExpressionMapper):
    def map_subscript(self, expr, type_context):
        from dune.perftool.generation.loopy import UnmanagedTemporaryVariable
        if isinstance(self.find_array(expr), UnmanagedTemporaryVariable):
            # If there is but one index, we do not need to handle this
            from pymbolic.primitives import Subscript, Variable
            if isinstance(expr.index, (Variable, int)):
                return expr

            # Else, we construct a nested Subscript chain
            ret = expr.aggregate
            for i in expr.index:
                ret = Subscript(ret, i)
            return ret
        else:
            return ExpressionToCExpressionMapper.map_subscript(self, expr, type_context)


class DuneASTBuilder(CASTBuilder):
    def get_expression_to_c_expression_mapper(self, codegen_state):
        return MyMapper(codegen_state)

    def get_temporary_decls(self, codegen_state, schedule_index):
        # Currently we do *not* want to build the temporary declarations
        # through loopy, as this would involve lots of work on a proper
        # type system!
        return []


class DuneTarget(TargetBase):
    def __init__(self):
        # Set fortran_abi to allow reusing CASTBuilder for the moment
        self.fortran_abi = False

    def split_kernel_at_global_barriers(self):
        return False

    def get_host_ast_builder(self):
        return DummyHostASTBuilder(self)

    def get_device_ast_builder(self):
        return DuneASTBuilder(self)

    def dtype_to_typename(self, dtype):
        # For now, we do this the simplest possible way
        return _registry[dtype.dtype.name]

    def is_vector_dtype(self, dtype):
        return False
